import os
import re
import sys
import copy
import shutil
forge = sys.modules["forge"]

# NOTE: The Maya version of boost may be different.
# The boost-regex appears to be sensitive and they don't provide the dso.
# Moving aside their include/boost seems to fix regex,
# but it leaves their other boost dso's cross-versioned with system includes.

def setup(module):
	module.cppmap = {}
	test_include = {}
	test_cpp = {}
	test_link = {}
	test_variant = forge.maya_sources[0][0]

	deplibs = forge.corelibs+ [
				"fexDrawDLLib",
				"fexMetaDLLib",
				"fexOperateDLLib",
				"fexSurfaceDLLib",
				"fexWindowLib"	]

	fe_maya_oplibs = ':'.join(forge.maya_oplibs_replace)
	if fe_maya_oplibs == '':
		if os.path.exists('ext/grass'):
			fe_maya_oplibs += ':fexGrassDL'
			sys.stdout.write("grass ")
		if os.path.exists('ext/vegetation'):
			fe_maya_oplibs += ':fexVegetationDL'
			sys.stdout.write("vegetation ")
	else:
		fe_maya_oplibs = ':' + fe_maya_oplibs
	if len(forge.maya_oplibs_append):
		fe_maya_oplibs += ':' + ':'.join(forge.maya_oplibs_append)

	for maya_source in forge.maya_sources:
		if not os.path.exists(maya_source[1]) or not os.path.exists(maya_source[2]):
			forge.color_on(0, forge.BLUE)
			sys.stdout.write("-" + maya_source[0] + " ")
			forge.color_off()
			continue

		if maya_source[0] == "":
			sys.stdout.write(". ")
		else:
			sys.stdout.write(maya_source[0] + " ")

		srcList = [	"maya.pmh",
					"CurveCombineOp",
					"CurveSeparateOp",
					"JointCombineOp",
					"JointSeparateOp",
					"MayaBrush",
					"MayaDraw",
					"MayaGraph",
					"MayaNode",
					"SurfaceAccessibleMaya",
					"SurfaceAccessorMaya",
					"SurfaceCurvesMaya",
					"SurfaceTrianglesMaya",
					"assertMaya",
					"mayaDL" ]

		variant = maya_source[0]
		if variant != test_variant:
			variantPath = module.modPath + '/' + variant
			if os.path.lexists(variantPath) == 0:
				os.symlink('.', variantPath)

			srcListVariant = []
			for src in srcList:
				srcListVariant += [ variant + '/' + src ]
			srcList = srcListVariant

		variantSplit = maya_source[1].split('/')
		variantSplit = variantSplit[len(variantSplit)-2].split('-')
		maya_major = variantSplit[0].replace('maya', '')

		dllname = "fexMayaDL" + variant

		dll = module.DLL( dllname, srcList )

		fe_maya_libs = dllname + fe_maya_oplibs

		for src in srcList:
			relativePath = forge.RelativePath(forge.rootPath,module.modPath)
			modObjPath = os.path.join(forge.objPath, relativePath)
			objName = modObjPath + '/' + src
			if re.compile(r'(.*)\.pmh$').match(src):
				objName += '.gch'
			else:
				objName += forge.objsuf
			srcTarget = forge.FindTarget(objName)
			srcTarget.includemap = { 'maya_src' : maya_source[1] }
			srcTarget.cppmap = { 'maya_src' :
					' -DFE_MAYA_MAJOR=' + str(maya_major) }
			srcTarget.cppmap['maya_src'] += ' -DFE_MAYA_LIBS=' + fe_maya_libs
			if variant == test_variant:
				test_include = copy.deepcopy(srcTarget.includemap)
				test_cpp = copy.deepcopy(srcTarget.cppmap)

		# this overrides previous -W flags in codegen
		module.cppmap = {}
		module.cppmap['maya'] = ' -Wno-deprecated -Wall -W -Wno-parentheses -Wno-sign-compare -Wno-reorder -Wno-uninitialized -Wunused -Wno-unused-parameter'

		dll.linkmap = {}
		if forge.fe_os == "FE_LINUX":
			dll.linkmap["mayalibs"] = "-Xlinker -rpath='$ORIGIN/../'"
			dll.linkmap["mayalibs"] += " -shared"
			dll.linkmap["mayalibs"] += " -L" + maya_source[2]
			dll.linkmap["mayalibs"] += " -lOpenMaya -lOpenMayaUI -lOpenMayaAnim"
		elif forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
			dll.linkmap["mayalibs"] = "TODO"

		if variant == test_variant:
			test_link = copy.deepcopy(dll.linkmap)

		forge.deps( [ dllname + "Lib" ], deplibs )

		# make sym links for maya to look at
		# TODO should work on non-linux
		mayaDso = 'lib' + dllname + '.so'

		source = os.path.join('..', mayaDso)

		mayaPath = os.path.join(forge.libPath, 'maya') + variant
		if os.path.isdir(mayaPath) == 0:
			os.makedirs(mayaPath, 0755)

		dest = os.path.join(mayaPath, 'libfexMayaDL.so')
		if os.path.lexists(dest) == 0:
			os.symlink(source, dest)

		# provoke RTLD_GLOBAL with empty 'libfexMayaDL.sog'
		sog = file(dest + 'g', "w")
		sog.close()

		mayaScriptSource = 'ext/maya/scripts/'

		mayaScriptPath = os.path.join(mayaPath, 'scripts')
		if os.path.isdir(mayaScriptPath) == 0:
			os.makedirs(mayaScriptPath, 0755)

		melFiles = os.listdir(mayaScriptSource)
		for mel in melFiles:
			melSource = os.path.join(mayaScriptSource, mel)
			melScript = os.path.join(mayaScriptPath, mel)
			shutil.copyfile(melSource, melScript)

		fe_local = 'fe_local.mel'
		if os.path.exists(fe_local):
			melScript = os.path.join(mayaScriptPath, fe_local)
			shutil.copyfile(fe_local, melScript)

		mayaIconPath = os.path.join(mayaPath, 'icons')
		if os.path.isdir(mayaIconPath) == 0:
			os.makedirs(mayaIconPath, 0755)

		for image in [
				'FE.svg',
				'FE_alpha.svg',
				'FE_beta.svg',
				'FE_maya.svg',
				'FE_240.svg',
				'FreeElectron.svg' ]:
			icon = os.path.join(mayaIconPath, image)
			shutil.copyfile('doc/image/' + image, icon)

		#* TODO automated list
		for image in [
				'BladeOp.svg',
				'DodgeOp.svg',
				'SpineFitOp.svg',
				'TwistBindOp.svg',
				'TwistWrapOp.svg',
				'CurveCombineOp.svg',
				'CurveSeparateOp.svg',
				'BloatOp.svg',
				'ClaspOp.svg',
				'ConnectOp.svg',
				'CurvaceousOp.svg',
				'CurveCreateOp.svg',
				'ExcarnateOp.svg',
				'FollicleOp.svg',
				'FusionOp.svg',
				'GridWrapOp.svg',
				'GroupOp.svg',
				'HammerOp.svg',
				'HobbleOp.svg',
				'ImportOp.svg',
				'JointCombineOp.svg',
				'JointSeparateOp.svg',
				'LengthCorrectOp.svg',
				'NexusOp.svg',
				'OffsetOp.svg',
				'PartitionOp.svg',
				'PuppetOp.svg',
				'RasterOp.svg',
				'SpreadsheetOp.svg',
				'SurfaceAttrCopyOp.svg',
				'SurfaceAttrCreateOp.svg',
				'SurfaceAttrLabOp.svg',
				'SurfaceBindOp.svg',
				'SurfaceCopyOp.svg',
				'SurfaceMetricOp.svg',
				'SurfaceNormalOp.svg',
				'SurfaceSampleOp.svg',
				'SurfaceSummaryOp.svg',
				'SurfaceWrapOp.svg',
				'XRayOp.svg']:
			dest = os.path.join(mayaIconPath, image)
			if os.path.lexists(dest) == 0:
				os.symlink('FE_240.svg', dest)

	test = module.Module('test')
	test.includemap = test_include
	test.cppmap = {}
	test.cppmap = copy.deepcopy(module.cppmap)
	if 'maya_src' in test_cpp:
		test.cppmap['maya'] += test_cpp['maya_src']
	test.linkmap = copy.deepcopy(test_link)
	test_deplibs =	forge.corelibs + [
				"fexMayaDL" + test_variant + "Lib" ]

	tests = [ 'xMayaContext' ]
	for t in tests:
		test_exe = test.Exe(t)
		forge.deps([t + "Exe"], test_deplibs)

def auto(module):
	test_file = """
#define _BOOL
#include <maya/MTypes.h>

int main(void)
{
	return(0);
}
    \n"""

	maya_path = os.getenv('MAYA_ROOT')
	if maya_path:
		forge.includemap['maya'] = os.path.join(maya_path,'include')

	result = forge.cctest(test_file)

	forge.includemap['maya'] = ''

	return result
