/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SGO_DEBUG		FALSE

#include "operator/operator.pmh"

using namespace fe;

void HingeOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* page Config

	catalog<I32>("faceA")=0;
	catalog<I32>("faceA","high")=100;
	catalog<I32>("faceA","max")=INT_MAX;
	catalog<String>("faceA","label")="Face A";
	catalog<bool>("faceA","joined")=true;
	catalog<String>("faceA","page")="Config";

	catalog<SpatialVector>("baryA")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("baryA","label")="Barycenter A";
	catalog<String>("baryA","page")="Config";

	catalog<SpatialVector>("offsetA");
	catalog<String>("offsetA","label")="Offset A";
	catalog<String>("offsetA","page")="Config";

	catalog<I32>("faceB")=0;
	catalog<I32>("faceB","high")=100;
	catalog<I32>("faceB","max")=INT_MAX;
	catalog<String>("faceB","label")="Face B";
	catalog<bool>("faceB","joined")=true;
	catalog<String>("faceB","page")="Config";

	catalog<SpatialVector>("baryB")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("baryB","label")="Barycenter B";
	catalog<String>("baryB","page")="Config";

	catalog<SpatialVector>("offsetB");
	catalog<String>("offsetB","label")="Offset B";
	catalog<String>("offsetB","page")="Config";

	//* page Behavior

	catalog<Real>("deadRadius")=0.0;
	catalog<Real>("deadRadius","high")=10.0;
	catalog<Real>("deadRadius","max")=1e6;
	catalog<String>("deadRadius","label")="Dead Radius";
	catalog<bool>("deadRadius","joined")=true;
	catalog<String>("deadRadius","page")="Behavior";

	catalog<Real>("liveRadius")=0.0;
	catalog<Real>("liveRadius","high")=10.0;
	catalog<Real>("liveRadius","max")=1e6;
	catalog<String>("liveRadius","label")="Live Radius";
	catalog<String>("liveRadius","page")="Behavior";

	catalog<Real>("threshold")=0.0;
	catalog<Real>("threshold","min")= -1e6;
	catalog<Real>("threshold","low")=0.0;
	catalog<Real>("threshold","high")=1.0;
	catalog<Real>("threshold","max")=1e6;
	catalog<String>("threshold","label")="Threshold";
	catalog<String>("threshold","page")="Behavior";

	catalog<bool>("soft")=false;
	catalog<String>("soft","label")="Soft";
	catalog<bool>("soft","joined")=true;
	catalog<String>("soft","page")="Behavior";

	catalog<Real>("softness")=0.0;
	catalog<Real>("softness","high")=1.0;
	catalog<Real>("softness","max")=1e6;
	catalog<String>("softness","label")="Softness";
	catalog<String>("softness","page")="Behavior";

	catalog<I32>("passes")=1;
	catalog<I32>("passes","min")=1;
	catalog<String>("passes","label")="Passes";
	catalog<String>("passes","page")="Behavior";
	catalog<bool>("passes","joined")=true;

	catalog<Real>("rate")=1.0;
	catalog<Real>("rate","max")=1.0;
	catalog<String>("rate","label")="Rate";
	catalog<String>("rate","page")="Behavior";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Collider Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;
}

void HingeOp::handle(Record& a_rSignal)
{
#if FE_SGO_DEBUG
	feLog("HingeOp::handle\n");
#endif

	const Real threshold=catalog<Real>("threshold");
	const BWORD soft=catalog<bool>("soft");
	const Real softness=catalog<Real>("softness");
	const Real deadRadius=catalog<Real>("deadRadius");
	const Real liveRadius=catalog<Real>("liveRadius");
	const I32 passes=catalog<I32>("passes");
	const Real rate=catalog<Real>("rate");

	sp<DrawI> spDrawGuide;
	accessGuide(spDrawGuide,a_rSignal,e_warning);

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputPoint;
	if(!access(spInputPoint,spInputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceI> spCollider;
	if(!access(spCollider,"Collider Surface")) return;

	spCollider->setRefinement(0);

	sp<SurfaceI> spDriver;
	access(spDriver,"Driver Surface",e_quiet);

	SpatialVector pivotA=catalog<SpatialVector>("offsetA");
	SpatialVector pivotB=catalog<SpatialVector>("offsetB");
	if(spDriver.isValid())
	{
		const I32 triangleIndexA=catalog<I32>("faceA");
		const SpatialBary baryA=catalog<SpatialVector>("baryA");
		const SpatialTransform xformA=spDriver->sample(triangleIndexA,baryA);
		transformVector(xformA,pivotA,pivotA);

		const I32 triangleIndexB=catalog<I32>("faceB");
		const SpatialBary baryB=catalog<SpatialVector>("baryB");
		const SpatialTransform xformB=spDriver->sample(triangleIndexB,baryB);
		transformVector(xformB,pivotB,pivotB);
	}

	if(spDrawGuide.isValid())
	{
//		sp<DrawableI> spDrawable(spCollider);
//		if(spDrawable.isValid())
//		{
//			spDrawable->draw(spDrawGuide,NULL);
//		}

		const Color green(0.0,1.0,0.0);
		const Color blue(0.0,0.0,1.0);

		Color colors[2];
		colors[0]=green;
		colors[1]=blue;

		SpatialVector line[2];
		line[0]=pivotA;
		line[1]=pivotB;

		spDrawGuide->drawLines(line,NULL,2,DrawI::e_strip,TRUE,colors);
	}

	const SpatialVector pivot=0.5*(pivotA+pivotB);
	const SpatialVector axis=unitSafe(pivotB-pivotA);

	const U32 pointCount=spOutputPoint->count();

	Real pointAngle[pointCount];

	for(I32 pass=0;pass<passes;pass++)
	{
		Real maxAngle=0.0;

		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			pointAngle[pointIndex]=0.0;

			const SpatialVector point=spOutputPoint->spatialVector(pointIndex);

#if FE_SGO_DEBUG
			feLog("\npoint %d/%d %s\n",pointIndex,pointCount,c_print(point));
#endif

			const SpatialVector from=point-pivot;
			const Real radius=magnitude(from);

			if(radius<=0.0 || radius<=deadRadius)
			{
				continue;
			}

			const SpatialVector fromUnit=from*(1.0/radius);
			const SpatialVector fromSide=cross(axis,fromUnit);

			const Real maxDistance= -1.0;
			const SpatialVector rayDir=unitSafe(fromSide);

#if FE_SGO_DEBUG
			feLog("rayDir %s\n",c_print(rayDir));
#endif

			const SpatialVector rayOrigin=point-threshold*rayDir;

			const U32 hitLimit=8;
			Array< sp<SurfaceI::ImpactI> > impactArray=
					spCollider->rayImpacts(rayOrigin,rayDir,
					maxDistance,hitLimit);

			const U32 hitCount=impactArray.size();
			if(hitCount)
			{
				sp<SurfaceI::ImpactI> spImpact=impactArray[hitCount-1];

				const SpatialVector intersection=spImpact->intersection();

#if FE_SGO_DEBUG
				const SpatialVector hitNormal=spImpact->normal();
				feLog("intersect %s normal %s\n",
						c_print(intersection),c_print(hitNormal));
#endif

				const SpatialVector goal=intersection+threshold*rayDir;
				const SpatialVector toUnit=unitSafe(goal-pivot);

				//* TODO not always fully locked to axis
				const SpatialVector fromLocked=unitSafe(cross(fromSide,axis));
				const SpatialVector toSide=cross(axis,toUnit);
				const SpatialVector toLocked=unitSafe(cross(toSide,axis));

#if FE_SGO_DEBUG
				feLog("from %s to %s\n",
						c_print(fromLocked),c_print(toLocked));
#endif

				SpatialQuaternion rotation;
				set(rotation,fromLocked,toLocked);

				Real rotAngle=0.0;
				SpatialVector rotAxis;
				rotation.computeAngleAxis(rotAngle,rotAxis);

				const Real dotAxis=dot(axis,rotAxis);
				if(dotAxis<0.0)
				{
					rotAngle= -rotAngle;
				}

				Real fade=1.0;
				if(radius<liveRadius && liveRadius>deadRadius)
				{
					fade=(radius-deadRadius)/(liveRadius-deadRadius);
				}

#if FE_SGO_DEBUG
				feLog("rotation %s\n",c_print(rotation));
				feLog("angle %.6G axis %s fade %.6G\n",
						rotAngle,c_print(rotAxis),fade);
#endif

				rotAngle*=fade;

				if(maxAngle<rotAngle)
				{
					maxAngle=rotAngle;
				}

				pointAngle[pointIndex]=rotAngle;

				if(spDrawGuide.isValid())
				{
					const Color blue(0.0,0.0,1.0);
					const Color green(0.0,1.0,0.0);
					const Color yellow(1.0,1.0,0.0);
					const Color white(1.0,1.0,1.0);

					Color colors[2];
					colors[0]=blue;
					colors[1]=green;

					SpatialVector line[2];
					line[0]=point;
					line[1]=point+fromLocked;

					spDrawGuide->drawLines(line,NULL,2,
							DrawI::e_strip,TRUE,colors);

					colors[1]=yellow;
					line[1]=point+toLocked;

					spDrawGuide->drawLines(line,NULL,2,
							DrawI::e_strip,TRUE,colors);

					colors[0]=white;
					line[0]=goal;

					spDrawGuide->drawLines(line,NULL,2,
							DrawI::e_strip,TRUE,colors);
				}
			}

			if(spDrawGuide.isValid())
			{
				const Color blue(0.0,0.0,1.0);
				const Color lightred(1.0,0.3,0.3);

				Color colors[2];
				colors[0]=blue;
				colors[1]=lightred;

				SpatialVector line[2];
				line[0]=point;
				line[1]=point+rayDir*radius*pointAngle[pointIndex];

				spDrawGuide->drawLines(line,NULL,2,
						DrawI::e_strip,TRUE,colors);
			}
		}

		if(soft)
		{
			const U32 primitiveCount=spOutputVertices->count();

			while(TRUE)
			{
				BWORD changed=FALSE;

				for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
						primitiveIndex++)
				{
					const U32 subCount=
							spOutputVertices->subCount(primitiveIndex);
					for(U32 subIndex=0;subIndex<subCount;subIndex++)
					{
						//* TODO cache lookups

						const U32 subIndexB=(subIndex+1)%subCount;

						const U32 pointIndexA=
								spOutputVertices->integer(
								primitiveIndex,subIndex);
						const U32 pointIndexB=
								spOutputVertices->integer(
								primitiveIndex,subIndexB);

						Real& rAngleA=pointAngle[pointIndexA];
						Real& rAngleB=pointAngle[pointIndexB];

						if(rAngleA<1e-6 && rAngleB<1e-6)
						{
							continue;
						}

						const SpatialVector pointA=
								spInputPoint->spatialVector(pointIndexA);
						const SpatialVector pointB=
								spInputPoint->spatialVector(pointIndexB);

						const Real distance=magnitude(pointB-pointA);
						if(distance<1e-6)
						{
							continue;
						}

						const Real limit=softness*distance;

//						feLog("vert %d:%d point %d %d angle %.6G %.6G"
//								" limit %.6G\n",
//								primitiveIndex,subIndex,
//								pointIndexA,pointIndexB,
//								rAngleA,rAngleB,limit);

//						const Real minA=rAngleB-limit;
						const Real minA=rAngleB*(1.0-limit);
						if(rAngleA<minA)
						{
							rAngleA=minA;
							changed=TRUE;
						}
						else
						{
//							const Real minB=rAngleA-limit;
							const Real minB=rAngleA*(1.0-limit);
							if(rAngleB<minB)
							{
								rAngleB=minB;
								changed=TRUE;
							}
						}
					}
				}

				if(!changed)
				{
					break;
				}
			}
		}

		if(maxAngle>0.0)
		{
			SpatialTransform conversion;

			if(!soft)
			{
				calcConversion(conversion,rate*maxAngle,axis,pivot);
			}

#if FE_SGO_DEBUG
			feLog("conversion\n%s\n",c_print(conversion));
#endif

			//* rotate all points
			for(U32 otherIndex=0;otherIndex<pointCount;otherIndex++)
			{
				if(soft)
				{
					calcConversion(conversion,
							rate*pointAngle[otherIndex],axis,pivot);
				}

				const SpatialVector otherPoint=
						spOutputPoint->spatialVector(otherIndex);

				SpatialVector transformed;
				transformVector(conversion,otherPoint,transformed);

				spOutputPoint->set(otherIndex,transformed);
			}
		}
	}

#if FE_SGO_DEBUG
	feLog("HingeOp::handle done\n");
#endif
}

//* static
void HingeOp::calcConversion(SpatialTransform& a_rTransform,
	const Real a_angle,const SpatialVector& a_rAxis,
	const SpatialVector& a_rPivot)
{
	setIdentity(a_rTransform);

	translate(a_rTransform,a_rPivot);

	const SpatialQuaternion rotation(a_angle,a_rAxis);
	a_rTransform*=SpatialTransform(rotation);

	translate(a_rTransform,-a_rPivot);
}
