/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "platform.pmh"

#define FE_BACKTRACE_GDB	(FE_CODEGEN<=FE_DEBUG)
#define FE_BATCH_GDB		FALSE

#if FE_COMPILER==FE_GNU
#include <cxxabi.h>
#if FE_OS!=FE_OSX
#include <execinfo.h>
#endif
#define GNU_SOURCE
#include <dlfcn.h>

void feLogBacktrace(U32 skip,BWORD debug)
{
#if FE_OS!=FE_OSX
	fe::String value;
	if(!fe::System::getEnvironmentVariable("FE_E_BACKTRACE", value) ||
			!atoi(value.c_str()))
	{
		return;
	}

	const U32 backtrace_size = 256;
	const U32 buffer_length=1024;
	const I32 skip_bottom=2;	// before main()

	void* stack[backtrace_size];
	I32 count=backtrace(stack,backtrace_size);
	char** symbols=backtrace_symbols(stack,count);

	feLogDirect("[35mBACKTRACE[0m\n");
	feLogDirect("Backtrace:\n");

//	backtrace_symbols_fd(stack, count, 2);

	Dl_info info;
	for(I32 m=skip;m<count-skip_bottom;m++)
	{
		if(dladdr(stack[m],&info) != 0)
		{
			char sym_name[buffer_length];
			sym_name[0]=0;

			char* start=strchr(symbols[m],'(');
			if(start)
			{
				I32 symlen=strchr(start,')')-start;
				if(symlen>0)
				{
					strncpy(sym_name,start+1,symlen-1);
					sym_name[symlen-1]=0;
				}
			}

			char buffer[buffer_length];
			FE_UWORD length=buffer_length;
			I32 status = 0;

			char* demangle=
					abi::__cxa_demangle(info.dli_sname,buffer,&length,&status);
/*
			feLogDirect("%s\n%s\n%s\n%s\n\n",
					demangle,
					sym_name,
					info.dli_fname,
					symbols[m]);
*/
			feLogDirect("%d: [37m[1m%s\n",
					m-skip,demangle? demangle: sym_name);
			feLogDirect("[0m[2m");

#if FALSE
			// doesn't work on DL's and not very accurate either
			if(system(NULL))
			{
				char command[buffer_length];
				sprintf(command, "addr2line -e %s %p",
						info.dli_fname,stack[m]);
				system(command);
			}
#endif
			feLogDirect("%s\n",info.dli_fname);
			feLogDirect("[0m");
		}
	}

	feLogDirect("[35mEND BACKTRACE[0m\n");

	if(symbols)
	{
		free(symbols);
	}

	if(!debug)
	{
		return;
	}

#if FE_BACKTRACE_GDB
	const BWORD attach=
			(fe::System::getEnvironmentVariable("FE_E_ATTACH", value) &&
				atoi(value.c_str()));

#if !FE_BATCH_GDB
	if(!attach)
	{
		return;
	}
#endif

	feLogDirect("[35mSTARTING GDB[0m\n");

	if(system(NULL) && dladdr(stack[count-1],&info) != 0)
	{
//		char* exename=info.dli_fname;
		char exename[buffer_length];
		exename[readlink("/proc/self/exe", exename, sizeof(exename)-1)]='\0';

		pid_t pid=getpid();
		char command[buffer_length];
		if(attach)
		{
			sprintf(command,"gdb -quiet %s %d",exename,pid);
		}
		else
		{
			sprintf(command,
					"echo 'bt\ndetach\nquit\n'| gdb -batch -x /dev/stdin %s %d",
					exename,pid);
		}
//		feLogDirect("Running: %s\n",command);
		int result=system(command);
		if(result)
		{
			feLogDirect("shell returned %d\n", result);
		}
	}

	feLogDirect("[35mEND GDB[0m\n");
#endif
#endif
}

#include <signal.h>

void feSegvHandler(int signal)
{
	feLogDirect("[31mSEGMENTATION FAULT[0m\n");
	feLogDirect("Free Electron has intercepted a crash"
			" using a registered handler.\n");
	feLogDirect("Note that any loaded code may be at fault.\n");
	feLogBacktrace();
	exit(139);
}

void feRegisterSegvHandler(void)
{
	struct sigaction action;
	memset(&action,0,sizeof(action));
	action.sa_handler=feSegvHandler;
	sigemptyset(&action.sa_mask);
	action.sa_flags=0;

	int fail=sigaction(SIGSEGV,&action,NULL);
	if(fail)
	{
		feLogDirect("feRegisterSegvHandler failed\n");
	}
}
#else
void feLogBacktrace(U32)
{
	feLogDirect("[35mBACKTRACE[0m: not implemented\n");
}

void feRegisterSegvHandler(void)
{
	feLogDirect("feRegisterSegvHandler not implemented\n");
}
#endif
